from proto_matcher.matcher.matcher import equals_proto
from proto_matcher.matcher.matcher import approximately
from proto_matcher.matcher.matcher import ignoring_field_paths
from proto_matcher.matcher.matcher import ignoring_repeated_field_ordering
from proto_matcher.matcher.matcher import partially



from typing import Optional, Set, Tuple, Union

from google.protobuf import message
from google.protobuf import text_format
from hamcrest.core.base_matcher import BaseMatcher
from hamcrest.core.description import Description
from hamcrest.core.helpers.wrap_matcher import wrap_matcher
from hamcrest.core.matcher import Matcher

from proto_matcher.compare import proto_compare, ProtoComparisonOptions
from proto_matcher.compare import ProtoComparisonScope
from proto_matcher.compare import ProtoFloatComparison
from proto_matcher.compare import RepeatedFieldComparison

_ProtoValue = Union[str, message.Message]

_ProtoMatcher = BaseMatcher[message.Message]


class _EqualsProto(_ProtoMatcher):

    def __init__(self, msg: _ProtoValue):
        self._msg = msg
        self._opts = ProtoComparisonOptions()

    def mut_options(self) -> ProtoComparisonOptions:
        return self._opts

    def matches(self,
                item: message.Message,
                mismatch_description: Optional[Description] = None) -> bool:
        expected = self._msg
        if isinstance(self._msg, str):
            proto_type = type(item)
            expected = text_format.Parse(self._msg, proto_type())
        cmp_result = proto_compare(item, expected, opts=self._opts)
        if not cmp_result.is_equal and mismatch_description:
            mismatch_description.append_text(cmp_result.explanation)
        return cmp_result.is_equal

    def describe_mismatch(self, item: message.Message,
                          mismatch_description: Description):
        self.matches(item, mismatch_description)

    def describe_to(self, description: Description):
        description.append_text(f"a protobuf of:\n{self._msg}")


def equals_proto(expected: message.Message) -> _ProtoMatcher:
    return _EqualsProto(expected)


def partially(matcher: _ProtoMatcher) -> _ProtoMatcher:
    matcher.mut_options().scope = ProtoComparisonScope.PARTIAL
    return matcher


def approximately(matcher: _ProtoMatcher,
                  float_margin: Optional[float] = None,
                  float_fraction: Optional[float] = None) -> _ProtoMatcher:
    opts = matcher.mut_options()
    opts.float_comp = ProtoFloatComparison.APPROXIMATE
    if float_margin:
        opts.float_margin = float_margin
    if float_fraction:
        opts.float_fraction = float_fraction
    return matcher


def ignoring_field_paths(field_paths: Set[Tuple[str]],
                         matcher: _ProtoMatcher) -> _ProtoMatcher:
    opts = matcher.mut_options()
    opts.ignore_field_paths = field_paths
    return matcher


def ignoring_repeated_field_ordering(matcher: _ProtoMatcher) -> _ProtoMatcher:
    opts = matcher.mut_options()
    opts.repeated_field_comp = RepeatedFieldComparison.AS_SET
    return matcher



from typing import Any, Callable, Iterable, Iterator, Optional, TypeVar, Tuple

T = TypeVar("T")

KeyFn = Callable[[T], Any]


def zip_pairs(
        xs: Iterable[T],
        ys: Iterable[T],
        key: Optional[KeyFn] = None
) -> Iterator[Tuple[Optional[T], Optional[T]]]:
    if not key:
        key = lambda x: 0

    xs = list(reversed(sorted(xs, key=key)))
    ys = list(reversed(sorted(ys, key=key)))

    while xs or ys:
        if not xs:
            yield None, ys.pop()
            continue
        if not ys:
            yield xs.pop(), None
            continue

        x_key = key(xs[-1])
        y_key = key(ys[-1])
        yield (xs.pop() if x_key <= y_key else None,
               ys.pop() if y_key <= x_key else None)



import unittest

from google.protobuf import text_format

from proto_matcher.compare import compare
from proto_matcher.testdata import test_pb2

_TEST_PROTO = """
bars {
    short_id: -123
    name: "a bar"
    size: 1
    notes: "hehe"
    notes: "123"
}
bars {
    long_id: 888899990000
    progress: 0.31415926
    checked: True
    notes: "photo"
}
baz {
    status: ERROR
}
mapping {
    key: 5
    value: "haha"
}
mapping {
    key: 10
    value: "hello world!"
}
"""


class ProtoCompareTest(unittest.TestCase):

    def assertProtoCompareToBe(self, result: compare.ProtoComparisonResult,
                               to_be: bool):
        self.assertEqual(result.is_equal, to_be, result.explanation)\

    def test_proto_comparable(self):
        self.assertTrue(compare.proto_comparable(test_pb2.Foo(),
                                                 test_pb2.Foo()))
        self.assertFalse(
            compare.proto_comparable(test_pb2.Foo(), test_pb2.Bar()))
        self.assertFalse(
            compare.proto_comparable(test_pb2.Baz(), test_pb2.Bar()))

        foo1 = test_pb2.Foo()
        foo1.baz.status = test_pb2.Baz.OK
        foo2 = test_pb2.Foo()
        bar = test_pb2.Bar()
        bar.progress = 0.75
        foo2.bars.append(bar)
        self.assertTrue(compare.proto_comparable(foo1, foo1))
        self.assertTrue(compare.proto_comparable(foo1, foo2))
        self.assertFalse(compare.proto_comparable(foo1, bar))

    def test_basic_equality(self):
        expected = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        actual = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        self.assertProtoCompareToBe(compare.proto_compare(actual, expected),
                                    True)

    def test_basic_inequality(self):
        expected = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        actual = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        expected.baz.Clear()
        self.assertProtoCompareToBe(compare.proto_compare(actual, expected),
                                    False)

    def test_repeated_field_inequality(self):
        expected = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        actual = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        expected.bars.add().progress = 0.1
        self.assertProtoCompareToBe(compare.proto_compare(actual, expected),
                                    False)

    def test_map_field_inequality(self):
        expected = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        actual = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        expected.mapping[15] = 'luck'
        self.assertProtoCompareToBe(compare.proto_compare(actual, expected),
                                    False)

    def test_basic_partial_equality(self):
        expected = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        actual = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        opts = compare.ProtoComparisonOptions(
            scope=compare.ProtoComparisonScope.PARTIAL)
        self.assertProtoCompareToBe(
            compare.proto_compare(actual, expected, opts=opts), True)

    def test_partial_equality_test_extra_field(self):
        expected = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        actual = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        expected.baz.Clear()

        opts = compare.ProtoComparisonOptions(
            scope=compare.ProtoComparisonScope.PARTIAL)
        self.assertProtoCompareToBe(
            compare.proto_compare(actual, expected, opts=opts), True)

    def test_basic_partial_inequality(self):
        expected = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        actual = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        expected.baz.status = test_pb2.Baz.OK

        opts = compare.ProtoComparisonOptions(
            scope=compare.ProtoComparisonScope.PARTIAL)
        self.assertProtoCompareToBe(
            compare.proto_compare(actual, expected, opts=opts), False)

    def test_partial_inequality_missing_field(self):
        expected = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        actual = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        actual.baz.Clear()

        opts = compare.ProtoComparisonOptions(
            scope=compare.ProtoComparisonScope.PARTIAL)
        self.assertProtoCompareToBe(
            compare.proto_compare(actual, expected, opts=opts), False)

    def test_repeated_field_partial_inequality(self):
        expected = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        actual = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        expected.bars.add().progress = 0.1

        opts = compare.ProtoComparisonOptions(
            scope=compare.ProtoComparisonScope.PARTIAL)
        self.assertProtoCompareToBe(
            compare.proto_compare(actual, expected, opts=opts), False)

    def test_aproximate_equality(self):
        expected = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        actual = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        self.assertProtoCompareToBe(compare.proto_compare(actual, expected),
                                    True)

    def test_aproximate_modified_equality(self):
        expected = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        actual = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        expected.bars[0].progress = 2.300005
        actual.bars[0].progress = 2.300006
        self.assertProtoCompareToBe(compare.proto_compare(actual, expected),
                                    False)

        opts = compare.ProtoComparisonOptions(
            float_comp=compare.ProtoFloatComparison.APPROXIMATE)
        self.assertProtoCompareToBe(
            compare.proto_compare(actual, expected, opts=opts), True)

    def test_aproximate_modified_equality_double(self):
        expected = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        actual = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        expected.bars[0].precision = 2.3 + 1.1e-15
        actual.bars[0].precision = 2.3 + 1.2e-15
        self.assertProtoCompareToBe(compare.proto_compare(actual, expected),
                                    False)

        opts = compare.ProtoComparisonOptions(
            float_comp=compare.ProtoFloatComparison.APPROXIMATE)
        self.assertProtoCompareToBe(
            compare.proto_compare(actual, expected, opts=opts), True)

    def test_within_fraction_or_margin_float(self):
        expected = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        actual = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        expected.bars[0].progress = 100.0
        actual.bars[0].progress = 109.9
        self.assertProtoCompareToBe(compare.proto_compare(actual, expected),
                                    False)

        # fraction and margin do not matter when |float_comp| is EXACT.
        opts = compare.ProtoComparisonOptions(float_fraction=0.0,
                                              float_margin=10.0)
        self.assertProtoCompareToBe(
            compare.proto_compare(actual, expected, opts=opts), False)

        opts = compare.ProtoComparisonOptions(
            float_fraction=0.0,
            float_margin=10.0,
            float_comp=compare.ProtoFloatComparison.APPROXIMATE)
        self.assertProtoCompareToBe(
            compare.proto_compare(actual, expected, opts=opts), True)

        opts = compare.ProtoComparisonOptions(
            float_fraction=0.2,
            float_margin=0.0,
            float_comp=compare.ProtoFloatComparison.APPROXIMATE)
        self.assertProtoCompareToBe(
            compare.proto_compare(actual, expected, opts=opts), True)

        opts = compare.ProtoComparisonOptions(
            float_fraction=0.01,
            float_margin=0.0,
            float_comp=compare.ProtoFloatComparison.APPROXIMATE)
        self.assertProtoCompareToBe(
            compare.proto_compare(actual, expected, opts=opts), False)

        opts = compare.ProtoComparisonOptions(
            float_fraction=0.10,
            float_margin=10.0,
            float_comp=compare.ProtoFloatComparison.APPROXIMATE)
        self.assertProtoCompareToBe(
            compare.proto_compare(actual, expected, opts=opts), True)

    def test_oneof_inequality(self):
        expected = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        actual = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        expected.bars[0].long_id = expected.bars[0].short_id
        self.assertProtoCompareToBe(compare.proto_compare(actual, expected),
                                    False)

    def test_compare_proto_ignoring_fields(self):
        # a: 1,      2,    3, 9, 4, 5, 7,   2
        # b:   9, 0, 2, 7, 3,    4, 5,   6, 2
        pass

    def test_ignore_field_single(self):
        expected = text_format.Parse('baz { status: ERROR }', test_pb2.Foo())
        actual = text_format.Parse('', test_pb2.Foo())
        self.assertProtoCompareToBe(compare.proto_compare(actual, expected),
                                    False)

        opts = compare.ProtoComparisonOptions(ignore_field_paths={('baz',)})
        self.assertProtoCompareToBe(
            compare.proto_compare(actual, expected, opts=opts), True)

    def test_ignore_field_repeated(self):
        expected = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        actual = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        del actual.bars[:]
        self.assertProtoCompareToBe(compare.proto_compare(actual, expected),
                                    False)

        opts = compare.ProtoComparisonOptions(ignore_field_paths={('bars',)})
        self.assertProtoCompareToBe(
            compare.proto_compare(actual, expected, opts=opts), True)

    def test_ignore_field_multiple(self):
        expected = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        actual = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        del actual.bars[:]
        actual.baz.status = test_pb2.Baz.OK

        opts = compare.ProtoComparisonOptions(ignore_field_paths={('bars',)})
        self.assertProtoCompareToBe(
            compare.proto_compare(actual, expected, opts=opts), False)
        opts = compare.ProtoComparisonOptions(ignore_field_paths={('baz',)})
        self.assertProtoCompareToBe(
            compare.proto_compare(actual, expected, opts=opts), False)

        opts = compare.ProtoComparisonOptions(
            ignore_field_paths={('bars',), ('baz',)})
        self.assertProtoCompareToBe(
            compare.proto_compare(actual, expected, opts=opts), True)

    def test_ignore_field_nested(self):
        expected = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        actual = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        actual.bars[0].size = 2
        self.assertProtoCompareToBe(compare.proto_compare(actual, expected),
                                    False)

        opts = compare.ProtoComparisonOptions(ignore_field_paths={('bars',
                                                                   'size')})
        self.assertProtoCompareToBe(
            compare.proto_compare(actual, expected, opts=opts), True)

    def test_compare_proto_repeated_fields_ignoring_order(self):
        expected = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        actual = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        reversed_bars = actual.bars[::-1]
        del actual.bars[:]
        actual.bars.extend(reversed_bars)
        self.assertProtoCompareToBe(compare.proto_compare(actual, expected),
                                    False)

        opts = compare.ProtoComparisonOptions(
            repeated_field_comp=compare.RepeatedFieldComparison.AS_SET)
        self.assertProtoCompareToBe(
            compare.proto_compare(actual, expected, opts=opts), True)

    def test_compare_proto_repeated_fields_ignoring_order__does_not_modify_input(self):
        expected = text_format.Parse(_TEST_PROTO, test_pb2.Foo())
        actual = text_format.Parse(_TEST_PROTO, test_pb2.Foo())

        opts = compare.ProtoComparisonOptions(
            repeated_field_comp=compare.RepeatedFieldComparison.AS_SET)
        compare.proto_compare(actual, expected, opts=opts)

        self.assertEqual(
            expected, text_format.Parse(_TEST_PROTO, test_pb2.Foo()))
        self.assertEqual(actual, text_format.Parse(_TEST_PROTO, test_pb2.Foo()))

    def test_ignore_nested_field_with_ignore_repeated_field_order(self):
        expected = test_pb2.Foo()
        expected.bars.extend([
            test_pb2.Bar(
                short_id=1,
                name='first bar',
            ),
            test_pb2.Bar(
                short_id=2,
                name='second bar',
            ),
        ])
        actual = test_pb2.Foo()
        actual.bars.extend([
            test_pb2.Bar(
                long_id=20,
                name='second bar',
            ),
            test_pb2.Bar(
                long_id=10,
                name='first bar',
            ),
        ])

        self.assertProtoCompareToBe(compare.proto_compare(actual, expected),
                                    False)
        opts = compare.ProtoComparisonOptions(
            repeated_field_comp=compare.RepeatedFieldComparison.AS_SET)
        self.assertProtoCompareToBe(
            compare.proto_compare(actual, expected, opts=opts), False)
        opts = compare.ProtoComparisonOptions(
            ignore_field_paths={('bars', 'short_id'), ('bars', 'long_id')})
        self.assertProtoCompareToBe(
            compare.proto_compare(actual, expected, opts=opts), False)

        opts = compare.ProtoComparisonOptions(
            repeated_field_comp=compare.RepeatedFieldComparison.AS_SET,
            ignore_field_paths={('bars', 'short_id'), ('bars', 'long_id')})
        self.assertProtoCompareToBe(
            compare.proto_compare(actual, expected, opts=opts), True)


if __name__ == '__main__':
    unittest.main()



import collections
import copy
import dataclasses
import enum
import math
import sys
from typing import Any, Generic, Iterable, List, Mapping, Optional, Set, TypeVar, Tuple

from google.protobuf import descriptor
from google.protobuf import message

from proto_matcher.compare import iter_util

_FieldDescriptor = descriptor.FieldDescriptor
_FLT_EPSILON = 1.19209e-07
_DBL_EPSILON = sys.float_info.epsilon


class RepeatedFieldComparison(enum.Enum):
    AS_LIST = enum.auto()
    AS_SET = enum.auto()


class ProtoComparisonScope(enum.Enum):
    FULL = enum.auto()
    PARTIAL = enum.auto()


class ProtoFloatComparison(enum.Enum):
    EXACT = enum.auto()
    APPROXIMATE = enum.auto()


@dataclasses.dataclass
class ProtoComparisonOptions:
    repeated_field_comp: RepeatedFieldComparison = RepeatedFieldComparison.AS_LIST
    scope: ProtoComparisonScope = ProtoComparisonScope.FULL
    ignore_field_paths: Optional[Set[Tuple[str]]] = None
    treating_nan_as_equal: bool = False
    float_comp: ProtoFloatComparison = ProtoFloatComparison.EXACT
    # |float_margin| and |float_fraction| are only used when
    # float_comp = APPROXIMATE.
    float_margin: Optional[float] = None
    float_fraction: Optional[float] = None


@dataclasses.dataclass
class ProtoComparisonResult:
    is_equal: bool = True
    explanation: str = ''


def proto_compare(actual: message.Message,
                  expected: message.Message,
                  opts: ProtoComparisonOptions = None) -> ProtoComparisonResult:
    if not proto_comparable(actual, expected):
        return ProtoComparisonResult(
            is_equal=False,
            explanation=
            f'Expected message of type: {expected.DESCRIPTOR.full_name}.'
            f'Actual: {actual.DESCRIPTOR.full_name}',
        )

    if not opts:
        opts = ProtoComparisonOptions()

    differencer = MessageDifferencer(opts, actual.DESCRIPTOR)
    # It's important for 'expected' to be the first argument here, as
    # compare() is not symmetric.  When we do a partial comparison,
    # only fields present in the first argument of compare() are
    # considered.
    return differencer.compare(expected, actual)


def proto_comparable(actual: message.Message,
                     expected: message.Message) -> bool:
    return actual.DESCRIPTOR == expected.DESCRIPTOR


T = TypeVar("T")


@dataclasses.dataclass
class ProtoFieldComparisonArgs(Generic[T]):
    expected: T
    actual: T
    field_desc: _FieldDescriptor
    field_path: Tuple[str]


class MessageDifferencer():

    def __init__(self, opts: ProtoComparisonOptions,
                 desc: descriptor.Descriptor):
        self._opts = opts
        if not self._opts.ignore_field_paths:
            self._opts.ignore_field_paths = set()
        # should expand ignored field paths using desc...
        self._desc = desc

    def compare(
        self,
        expected: message.Message,
        actual: message.Message,
        field_path: Tuple[str] = ()
    ) -> ProtoComparisonResult:
        return _combine_results([
            self._compare(
                ProtoFieldComparisonArgs(expected=expected,
                                         actual=actual,
                                         field_desc=None,
                                         field_path=field_path))
        ])

    def _compare(
        self, args: ProtoFieldComparisonArgs[message.Message]
    ) -> ProtoComparisonResult:
        return _combine_results([
            self._compare_field(
                ProtoFieldComparisonArgs(expected=args.expected,
                                         actual=args.actual,
                                         field_desc=field_desc,
                                         field_path=args.field_path))
            for field_desc in args.actual.DESCRIPTOR.fields
        ])

    def _compare_field(
            self, args: ProtoFieldComparisonArgs[Any]) -> ProtoComparisonResult:
        cmp_args = copy.copy(args)
        field_name = cmp_args.field_desc.name
        cmp_args.field_path = cmp_args.field_path + (field_name,)
        if cmp_args.field_path in self._opts.ignore_field_paths:
            return _equality_result()

        cmp_args.expected = getattr(cmp_args.expected, field_name, None)
        cmp_args.actual = getattr(cmp_args.actual, field_name, None)

        # Repeated field
        if cmp_args.field_desc.label == _FieldDescriptor.LABEL_REPEATED:
            # Map field
            if isinstance(cmp_args.expected, collections.abc.Mapping):
                return self._compare_map(cmp_args)
            return self._compare_repeated_field(cmp_args)

        # Singular field
        if (self._opts.scope == ProtoComparisonScope.PARTIAL and
                not _is_field_set(cmp_args.expected, cmp_args.field_desc)):
            return _equality_result()

        return self._compare_value(cmp_args)

    def _compare_repeated_field(
            self, cmp_args: ProtoFieldComparisonArgs[Iterable]
    ) -> ProtoComparisonResult:
        # Copy first to avoid modifying the original inputs.
        expected_list = list(cmp_args.expected)
        actual_list = list(cmp_args.actual)
        if self._opts.repeated_field_comp == RepeatedFieldComparison.AS_SET:
            # Identify as many matches as possible to minimize the number of
            # reported diffs.
            for expected in list(expected_list):
                for actual in list(actual_list):
                    item_result = self._compare_value(
                        ProtoFieldComparisonArgs(expected=expected,
                                                 actual=actual,
                                                 field_desc=cmp_args.field_desc,
                                                 field_path=cmp_args.field_path))
                    if item_result.is_equal:
                        actual_list.remove(actual)
                        expected_list.remove(expected)
                        break
            # If diffs remain, best-effort sort the lists to minimize the number
            # of diffs between each element.
            # (This is only best-effort because it fails to overlook ignored
            # fields.)
            as_set_key = lambda x: str(x)
            expected_list.sort(key=as_set_key)
            actual_list.sort(key=as_set_key)

        return _combine_results([
            self._compare_value(
                ProtoFieldComparisonArgs(expected=expected,
                                         actual=actual,
                                         field_desc=cmp_args.field_desc,
                                         field_path=cmp_args.field_path))
            for expected, actual in iter_util.zip_pairs(expected_list,
                                                        actual_list)
        ])

    def _compare_map(
            self, cmp_args: ProtoFieldComparisonArgs[Mapping]
    ) -> ProtoComparisonResult:
        desc = cmp_args.expected.GetEntryClass().DESCRIPTOR
        key_desc = desc.fields_by_name['key']
        value_desc = desc.fields_by_name['value']

        return _combine_results([
            _combine_results([
                self._compare_value(
                    ProtoFieldComparisonArgs(expected=expected_kv and
                                             expected_kv[0],
                                             actual=actual_kv and actual_kv[0],
                                             field_desc=key_desc,
                                             field_path=cmp_args.field_path)),
                self._compare_value(
                    ProtoFieldComparisonArgs(expected=expected_kv and
                                             expected_kv[1],
                                             actual=actual_kv and actual_kv[1],
                                             field_desc=value_desc,
                                             field_path=cmp_args.field_path))
            ]) for expected_kv, actual_kv in iter_util.zip_pairs(
                cmp_args.expected.items(),
                cmp_args.actual.items(),
                key=lambda kv: kv[0])
        ])

    def _compare_value(self, cmp_args: ProtoFieldComparisonArgs[Any]):
        if _is_message(cmp_args.field_desc):
            if cmp_args.expected and cmp_args.actual:
                return self._compare(cmp_args)
            return _inequality_result(cmp_args)
        if _is_float(cmp_args.field_desc):
            return self._compare_float(cmp_args)
        return _equality_result() if cmp_args.expected == cmp_args.actual \
            else _inequality_result(cmp_args)

    def _compare_float(
            self, cmp_args: ProtoFieldComparisonArgs) -> ProtoComparisonResult:
        if cmp_args.expected == cmp_args.actual:
            return _equality_result()
        if (self._opts.treating_nan_as_equal and math.isnan(expected) and
                math.isnan(actual)):
            return _equality_result()

        if self._opts.float_comp == ProtoFloatComparison.EXACT:
            return _inequality_result(cmp_args)

        # float_comp == APPROXIMATE
        fraction = self._opts.float_fraction or 0.0
        margin = self._opts.float_margin or _get_float_comparison_epsilon(
            cmp_args.field_desc)
        is_equal = _within_fraction_or_margin(cmp_args.expected,
                                              cmp_args.actual, fraction, margin)
        return _equality_result() if is_equal else _inequality_result(cmp_args)


def _combine_results(
        results: List[ProtoComparisonResult]) -> ProtoComparisonResult:
    return ProtoComparisonResult(
        is_equal=all([res.is_equal for res in results]),
        explanation='\n'.join(
            [res.explanation for res in results if res.explanation]),
    )


def _is_message(field_desc: _FieldDescriptor) -> bool:
    return field_desc.cpp_type == _FieldDescriptor.CPPTYPE_MESSAGE


def _is_float(field_desc: _FieldDescriptor) -> bool:
    return field_desc.cpp_type in (_FieldDescriptor.CPPTYPE_DOUBLE,
                                   _FieldDescriptor.CPPTYPE_FLOAT)


def _is_enum(field_desc: _FieldDescriptor) -> bool:
    return field_desc.enum_type is not None


def _is_field_set(value: Any, field_desc: _FieldDescriptor) -> bool:
    if _is_enum(field_desc):
        return value != 0
    return bool(value)


def _equality_result() -> ProtoComparisonResult:
    return ProtoComparisonResult()


def _inequality_result(
        cmp_args: ProtoFieldComparisonArgs) -> ProtoComparisonResult:
    return ProtoComparisonResult(
        is_equal=False,
        explanation=_explain_diff(cmp_args),
    )


def _explain_diff(cmp_args: ProtoFieldComparisonArgs):
    expected = _readable(cmp_args.expected, cmp_args.field_desc)
    actual = _readable(cmp_args.actual, cmp_args.field_desc)
    # TODO: add index here.
    field_path_with_index = '.'.join(cmp_args.field_path)
    if expected and not actual:
        return f'deleted: {field_path_with_index}: {expected}\n'
    if actual and not expected:
        return f'added: {field_path_with_index}: {actual}\n'
    return f'modified: {field_path_with_index}: {expected} -> {actual}\n'


def _readable(value: Any,
              value_desc: _FieldDescriptor,
              key_desc: Optional[_FieldDescriptor] = None) -> str:
    if key_desc and value:
        key, value = value
        return f'key: {_readable(key, key_desc)}' \
               f'value: {_readable(value, value_desc)}'
    if _is_enum(value_desc):
        return _get_enum_name(value, value_desc)
    if type(value) == str:
        return f'\"{value}\"'
    return str(value)


def _get_enum_name(enum_value: int, field_desc: _FieldDescriptor) -> str:
    return field_desc.enum_type.values[enum_value].name


def _get_float_comparison_epsilon(field_desc: _FieldDescriptor):
    if field_desc.cpp_type == _FieldDescriptor.CPPTYPE_DOUBLE:
        return _DBL_EPSILON * 32
    if field_desc.cpp_type == _FieldDescriptor.CPPTYPE_FLOAT:
        return _FLT_EPSILON * 32
    raise TypeError('Float comparison called on non-float types')


def _within_fraction_or_margin(x: float, y: float, fraction: float,
                               margin: float) -> bool:
    if not (fraction >= 0.0 and fraction < 1.0 and margin >= .0):
        raise ValueError(f'Invalid fraction {fraction} or margin {margin}')
    if math.isinf(x) or math.isinf(y):
        return False
    relative_margin = fraction * max(abs(x), abs(y))
    return abs(x - y) <= max(margin, relative_margin)



from proto_matcher.compare.compare import RepeatedFieldComparison
from proto_matcher.compare.compare import ProtoComparisonOptions
from proto_matcher.compare.compare import ProtoComparisonResult
from proto_matcher.compare.compare import ProtoComparisonScope
from proto_matcher.compare.compare import ProtoFloatComparison
from proto_matcher.compare.compare import proto_compare



import unittest

from hamcrest import *
from google.protobuf import text_format

from proto_matcher.matcher.matcher import equals_proto
from proto_matcher.matcher.matcher import approximately
from proto_matcher.matcher.matcher import ignoring_field_paths
from proto_matcher.matcher.matcher import ignoring_repeated_field_ordering
from proto_matcher.matcher.matcher import partially
from proto_matcher.compare import compare
from proto_matcher.testdata import test_pb2

_TEST_PROTO = """
bars {
    short_id: -123
    name: "a bar"
    size: 1
    notes: "hehe"
    notes: "123"
}
bars {
    long_id: 888899990000
    progress: 0.31415926
    checked: True
    notes: "photo"
}
baz {
    status: ERROR
}
mapping {
    key: 5
    value: "haha"
}
mapping {
    key: 10
    value: "hello world!"
}
"""


class ProtoCompareTest(unittest.TestCase):

    def assertProtoCompareToBe(self, result: compare.ProtoComparisonResult,
                               to_be: bool):
        self.assertEqual(result.is_equal, to_be, result.explanation)

    def _get_test_proto(self):
        return text_format.Parse(_TEST_PROTO, test_pb2.Foo())

    def test_basic_equality(self):
        expected = self._get_test_proto()
        assert_that(self._get_test_proto(), equals_proto(expected))

    def test_taking_expected_text_proto(self):
        assert_that(self._get_test_proto(), equals_proto(_TEST_PROTO))

    def test_basic_inequality(self):
        expected = self._get_test_proto()
        expected.baz.Clear()
        assert_that(self._get_test_proto(), not_(equals_proto(expected)))

    def test_repeated_field_inequality(self):
        expected = self._get_test_proto()
        expected.bars.add().progress = 0.1
        assert_that(self._get_test_proto(), not_(equals_proto(expected)))

    def test_map_field_inequality(self):
        expected = self._get_test_proto()
        expected.mapping[15] = 'luck'
        assert_that(self._get_test_proto(), not_(equals_proto(expected)))

    def test_basic_partial_equality(self):
        assert_that(self._get_test_proto(),
                    partially(equals_proto(_TEST_PROTO)))

    def test_partial_equality_test_extra_field(self):
        expected = self._get_test_proto()
        expected.baz.Clear()
        assert_that(self._get_test_proto(), partially(equals_proto(expected)))

    def test_basic_partial_inequality(self):
        expected = self._get_test_proto()
        expected.baz.status = test_pb2.Baz.OK
        assert_that(self._get_test_proto(),
                    not_(partially(equals_proto(expected))))

    def test_partial_inequality_missing_field(self):
        actual = self._get_test_proto()
        actual.baz.Clear()
        assert_that(actual, not_(partially(equals_proto(_TEST_PROTO))))

    def test_repeated_field_partial_inequality(self):
        expected = self._get_test_proto()
        expected.bars.add().progress = 0.1
        assert_that(self._get_test_proto(),
                    not_(partially(equals_proto(expected))))

    def test_aproximate_equality(self):
        actual = self._get_test_proto()
        assert_that(actual, approximately(equals_proto(_TEST_PROTO)))

    def test_aproximate_modified_equality(self):
        expected = self._get_test_proto()
        actual = self._get_test_proto()
        expected.bars[0].progress = 2.300005
        actual.bars[0].progress = 2.300006
        assert_that(actual, approximately(equals_proto(expected)))

    def test_aproximate_modified_equality_double(self):
        expected = self._get_test_proto()
        actual = self._get_test_proto()
        expected.bars[0].precision = 2.3 + 1.1e-15
        actual.bars[0].precision = 2.3 + 1.2e-15
        assert_that(actual, not_(equals_proto(expected)))
        assert_that(actual, approximately(equals_proto(expected)))

    def test_within_fraction_or_margin_float(self):
        expected = self._get_test_proto()
        actual = self._get_test_proto()
        expected.bars[0].progress = 100.0
        actual.bars[0].progress = 109.9
        assert_that(actual, not_(equals_proto(expected)))

        assert_that(actual,
                    approximately(equals_proto(expected), float_margin=10.0))
        assert_that(actual,
                    approximately(equals_proto(expected), float_fraction=0.2))
        assert_that(
            actual,
            not_(approximately(equals_proto(expected), float_fraction=0.01)))
        assert_that(
            actual,
            approximately(equals_proto(expected),
                          float_fraction=0.1,
                          float_margin=10.0))

    def test_oneof_inequality(self):
        expected = self._get_test_proto()
        expected.bars[0].long_id = expected.bars[0].short_id
        assert_that(self._get_test_proto(), not_(equals_proto(expected)))

    def test_compare_proto_ignoring_fields(self):
        # a: 1,      2,    3, 9, 4, 5, 7,   2
        # b:   9, 0, 2, 7, 3,    4, 5,   6, 2
        pass

    def test_ignore_field_single(self):
        expected = text_format.Parse('baz { status: ERROR }', test_pb2.Foo())
        actual = text_format.Parse('', test_pb2.Foo())
        assert_that(actual, not_(equals_proto(expected)))

        assert_that(actual,
                    ignoring_field_paths({('baz',)}, equals_proto(expected)))

    def test_ignore_field_repeated(self):
        expected = self._get_test_proto()
        actual = self._get_test_proto()
        del actual.bars[:]
        assert_that(actual, not_(equals_proto(expected)))

        assert_that(actual,
                    ignoring_field_paths({('bars',)}, equals_proto(expected)))

    def test_ignore_field_multiple(self):
        expected = self._get_test_proto()
        actual = self._get_test_proto()
        del actual.bars[:]
        actual.baz.status = test_pb2.Baz.OK

        assert_that(
            actual,
            not_(ignoring_field_paths({('baz',)}, equals_proto(expected))))
        assert_that(
            actual,
            not_(ignoring_field_paths({('bars',)}, equals_proto(expected))))
        assert_that(
            actual,
            ignoring_field_paths({('bars',), ('baz',)}, equals_proto(expected)))

    def test_ignore_field_nested(self):
        expected = self._get_test_proto()
        actual = self._get_test_proto()
        actual.bars[0].size = 2

        assert_that(actual, not_(equals_proto(expected)))

        assert_that(
            actual,
            ignoring_field_paths({('bars', 'size')}, equals_proto(expected)))

    def test_compare_proto_repeated_fields_ignoring_order(self):
        expected = self._get_test_proto()
        actual = self._get_test_proto()
        reversed_bars = actual.bars[::-1]
        del actual.bars[:]
        actual.bars.extend(reversed_bars)
        assert_that(actual, not_(equals_proto(expected)))

        assert_that(actual,
                    ignoring_repeated_field_ordering(equals_proto(expected)))

    def test_ignore_nested_field_with_ignore_repeated_field_order(self):
        expected = test_pb2.Foo()
        expected.bars.extend([
            test_pb2.Bar(
                short_id=1,
                name='first bar',
            ),
            test_pb2.Bar(
                short_id=2,
                name='second bar',
            ),
        ])
        actual = test_pb2.Foo()
        actual.bars.extend([
            test_pb2.Bar(
                long_id=20,
                name='second bar',
            ),
            test_pb2.Bar(
                long_id=10,
                name='first bar',
            ),
        ])

        assert_that(actual, not_(equals_proto(expected)))
        assert_that(
            actual,
            not_(ignoring_repeated_field_ordering(equals_proto(expected))))
        ignored_fields = {('bars', 'short_id'), ('bars', 'long_id')}
        assert_that(
            actual,
            not_(ignoring_field_paths(ignored_fields, equals_proto(expected))))
        assert_that(
            actual,
            ignoring_repeated_field_ordering(
                ignoring_field_paths(ignored_fields, equals_proto(expected))))
        assert_that(
            actual,
            ignoring_field_paths(
                ignored_fields,
                ignoring_repeated_field_ordering(equals_proto(expected))))


if __name__ == '__main__':
    unittest.main()
